@@ -27,7 +27,6 @@ The actual wheel builder.
 #
 
 # stdlib
-import configparser
 import os
 import pathlib
 import posixpath
@@ -36,22 +35,21 @@ import shutil
 import tarfile
 import warnings
 from abc import ABC, abstractmethod
+from datetime import datetime
 from email.headerregistry import Address
-from email.message import EmailMessage
 from functools import partial
-from io import StringIO
 from typing import Any, Dict, Iterator, Mapping, Optional
-from zipfile import ZipFile
 
 # 3rd party
 import click
 import dom_toml
-from consolekit.terminal_colours import Fore, resolve_color_default
+import handy_archives
+from consolekit.terminal_colours import ColourTrilean, Fore, resolve_color_default
+from dist_meta import entry_points, metadata, wheel
+from dist_meta.metadata_mapping import MetadataMapping
 from domdf_python_tools.paths import PathPlus, sort_paths, traverse_to_file
-from domdf_python_tools.stringlist import StringList
 from domdf_python_tools.typing import PathLike
 from domdf_python_tools.words import word_join
-from first import first
 from shippinglabel.checksum import get_record_entry
 from shippinglabel.requirements import ComparableRequirement, combine_requirements
 
@@ -96,9 +94,10 @@ class AbstractBuilder(ABC):
 			config: Mapping[str, Any],
 			build_dir: Optional[PathLike] = None,
 			out_dir: Optional[PathLike] = None,
-			*,
+			*args,
 			verbose: bool = False,
-			colour: bool = None,
+			colour: ColourTrilean = None,
+			**kwargs,
 			):
 
 		# Walk up the tree until a "pyproject.toml" file is found.
@@ -270,7 +269,7 @@ class AbstractBuilder(ABC):
 		:param \*entries:
 		"""
 
-		def copy_file(filename):
+		def copy_file(filename: PathPlus):
 			target = self.build_dir / filename.relative_to(self.project_dir / self.code_directory)
 			target.parent.maybe_make(parents=True)
 			shutil.copy2(src=filename, dst=target)
@@ -353,25 +352,15 @@ class AbstractBuilder(ABC):
 			target.write_clean(self.config["license"].text)
 			self.report_written(target)
 
-	def write_metadata(self, metadata_file: PathPlus):
+	def parse_authors(self) -> Dict[str, str]:
 		"""
-		Write `Core Metadata`_ to the given file.
+		Parse the :tconf:`project.authors` and :tconf:`~project.maintainers` fields into :core-meta:`Author`,
+		:core-meta:`Maintainer-Email` etc.
 
-		.. _Core Metadata: https://packaging.python.org/specifications/core-metadata
+		:return: A mapping of field names to values.
 
-		:param metadata_file:
-		"""
-
-		metadata = EmailMessage()
-
-		# TODO: metadata 2.2
-		# Need to translate pep621 dynamic into core metadata field names
-		metadata["Metadata-Version"] = "2.1"
-		metadata["Name"] = self.config["name"]
-		metadata["Version"] = str(self.config["version"])
-
-		if self.config["description"] is not None:
-			metadata["Summary"] = self.config["description"]
+			Possible field names are ``Author``, ``Author-Email``, ``Maintainer``, and ``Maintainer-Email``.
+		"""  # noqa: D400
 
 		author = []
 		author_email = []
@@ -398,45 +387,69 @@ class AbstractBuilder(ABC):
 
 		# TODO: I'm not quite sure how PEP 621 expects a name for one author and the email for another to be handled.
 
+		output = {}
+
 		if author_email:
-			metadata["Author-email"] = ", ".join(author_email)
+			output["Author-email"] = ", ".join(author_email)
 		elif author:
-			metadata["Author"] = word_join(author)
+			output["Author"] = word_join(author)
 
 		if maintainer_email:
-			metadata["Author-email"] = ", ".join(maintainer_email)
+			output["Maintainer-email"] = ", ".join(maintainer_email)
 		elif maintainer:
-			metadata["Author"] = word_join(maintainer)
+			output["Maintainer"] = word_join(maintainer)
 
-		if self.config["license-key"] is not None:
-			metadata["License"] = self.config["license-key"]
+		return output
+
+	def get_metadata_map(self) -> MetadataMapping:
+		"""
+		Generate the content of the ``METADATA`` / ``PKG-INFO`` file.
+		"""
+
+		metadata_mapping = MetadataMapping()
+
+		# TODO: metadata 2.2
+		# Need to translate pep621 dynamic into core metadata field names
+		metadata_mapping["Metadata-Version"] = "2.1"
+		metadata_mapping["Name"] = self.config["name"]
+		metadata_mapping["Version"] = str(self.config["version"])
+
+		def add_not_none(key: str, field: str):
+			if self.config[key] is not None:
+				metadata_mapping[field] = self.config[key]
+
+		def add_multiple(key: str, field: str):
+			for value in self.config[key]:
+				metadata_mapping[field] = str(value)
+
+		metadata_mapping.update(self.parse_authors())
+
+		add_not_none("description", "Summary")
+		add_not_none("license-key", "License")
+
+		add_multiple("classifiers", "Classifier")
+		add_multiple("dependencies", "Requires-Dist")
 
 		if self.config["keywords"]:
-			metadata["Keywords"] = ','.join(self.config["keywords"])
+			metadata_mapping["Keywords"] = ','.join(self.config["keywords"])
 
 		seen_hp = False
 
 		for category, url in self.config["urls"].items():
 			if category.lower() in {"homepage", "home page"} and not seen_hp:
-				metadata["Home-page"] = url
+				metadata_mapping["Home-page"] = url
 				seen_hp = True
 			else:
-				metadata["Project-URL"] = f"{category}, {url}"
+				metadata_mapping["Project-URL"] = f"{category}, {url}"
 
 		for platform in (self.config.get("platforms", None) or ()):
-			metadata["Platform"] = platform
-
-		for classifier in self.config["classifiers"]:
-			metadata["Classifier"] = classifier
+			metadata_mapping["Platform"] = platform
 
 		if self.config["requires-python"]:
-			metadata["Requires-Python"] = str(self.config["requires-python"])
-
-		for requirement in self.config["dependencies"]:
-			metadata["Requires-Dist"] = str(requirement)
+			metadata_mapping["Requires-Python"] = str(self.config["requires-python"])
 
 		for extra, requirements in self.config["optional-dependencies"].items():
-			metadata["Provides-Extra"] = extra
+			metadata_mapping["Provides-Extra"] = extra
 			for requirement in requirements:
 				requirement = ComparableRequirement(str(requirement))
 
@@ -445,24 +458,29 @@ class AbstractBuilder(ABC):
 				else:
 					requirement.marker = f"extra == {extra!r}"
 
-				metadata["Requires-Dist"] = str(requirement)
+				metadata_mapping["Requires-Dist"] = str(requirement)
 
 		# TODO:
 		#  https://packaging.python.org/specifications/core-metadata/#requires-external-multiple-use
 		#  https://packaging.python.org/specifications/core-metadata/#provides-dist-multiple-use
 		#  https://packaging.python.org/specifications/core-metadata/#obsoletes-dist-multiple-use
 
-		if self.config["readme"] is None:
-			description = ''
-		else:
-			metadata["Description-Content-Type"] = self.config["readme"].content_type
-			description = self.config["readme"].text
+		if self.config["readme"] is not None:
+			metadata_mapping["Description"] = self.config["readme"].text
+			metadata_mapping["Description-Content-Type"] = self.config["readme"].content_type
 
-		metadata_file.write_lines([
-				metadata.as_string(maxheaderlen=2048, policy=metadata.policy.clone(utf8=True)),
-				description,
-				])
+		return metadata_mapping
 
+	def write_metadata(self, metadata_file: PathPlus, metadata_mapping: MetadataMapping):
+		"""
+		Write `Core Metadata`_ to the given file.
+
+		.. _Core Metadata: https://packaging.python.org/specifications/core-metadata
+
+		:param metadata_file:
+		"""
+
+		metadata_file.write_text(metadata.dumps(metadata_mapping))
 		self.report_written(metadata_file)
 
 	def call_additional_hooks(self):
@@ -526,10 +544,8 @@ class SDistBuilder(AbstractBuilder):
 		with tarfile.open(sdist_filename, mode="w:gz", format=tarfile.PAX_FORMAT) as sdist_archive:
 			for file in self.build_dir.rglob('*'):
 				if file.is_file():
-					sdist_archive.add(
-							str(file),
-							arcname=posixpath.join(self.archive_name, file.relative_to(self.build_dir).as_posix()),
-							)
+					arcname = posixpath.join(self.archive_name, file.relative_to(self.build_dir).as_posix())
+					sdist_archive.add(str(file), arcname=arcname)
 
 		self._echo(Fore.GREEN(f"Source distribution created at {sdist_filename.resolve().as_posix()}"))
 		return os.path.basename(sdist_filename)
@@ -600,7 +616,7 @@ class SDistBuilder(AbstractBuilder):
 				self.report_copied(source, dest)
 
 		self.write_readme()
-		self.write_metadata(self.build_dir / "PKG-INFO")
+		self.write_metadata(self.build_dir / "PKG-INFO", self.get_metadata_map())
 		self.call_additional_hooks()
 
 		return self.create_sdist_archive()
@@ -687,34 +703,18 @@ class WheelBuilder(AbstractBuilder):
 		``[project.scripts]``, ``[project.gui-scripts]`` and ``[project.entry-points]``
 		"""  # noqa: D400
 
-		buf = StringList()
-		if self.config["scripts"]:
-			buf.append("[console_scripts]")
+		ep_dict = {}
 
-			for name, func in self.config["scripts"].items():
-				buf.append(f"{name} = {func}")
+		if self.config["scripts"]:
+			ep_dict["console_scripts"] = self.config["scripts"]
 
 		if self.config["gui-scripts"]:
-			buf.append("[gui_scripts]")
+			ep_dict["gui_scripts"] = self.config["gui-scripts"]
 
-			for name, func in self.config["gui-scripts"].items():
-				buf.append(f"{name} = {func}")
-
-		for group, entry_points in self.config["entry-points"].items():
-
-			buf.append(f"[{group}]")
-
-			for name, func in entry_points.items():
-				buf.append(f"{name} = {func}")
-
-		cfg_parser = configparser.ConfigParser()
-		cfg_parser.optionxform = str  # type: ignore  # preserve case
-		cfg_parser.read_string(str(buf))
-		cfg_io = StringIO()
-		cfg_parser.write(cfg_io)
+		ep_dict.update(self.config["entry-points"])
 
 		entry_points_file = self.dist_info / "entry_points.txt"
-		entry_points_file.write_clean(cfg_io.getvalue())
+		entry_points.dump(ep_dict, entry_points_file)
 		self.report_written(entry_points_file)
 
 	def write_wheel(self) -> None:
@@ -722,26 +722,37 @@ class WheelBuilder(AbstractBuilder):
 		Write the metadata to the ``WHEEL`` file.
 		"""
 
-		wheel = EmailMessage()
-		wheel["Wheel-Version"] = "1.0"
-		wheel["Generator"] = self.generator
-		wheel["Root-Is-Purelib"] = "true"
-		wheel["Tag"] = self.tag
-
 		wheel_file = self.dist_info / "WHEEL"
-		wheel_file.write_clean(str(wheel))
+
+		wheel_file.write_clean(
+				wheel.dumps({
+						"Wheel-Version": "1.0",
+						"Generator": self.generator,
+						"Root-Is-Purelib": True,
+						"Tag": [self.tag],
+						})
+				)
+
 		self.report_written(wheel_file)
 
-	def write_top_level(self, dist_info: PathPlus) -> None:
+	@staticmethod
+	def get_source_epoch() -> Optional[datetime]:
 		"""
-		Write the ``top_level.txt`` file to the ``*.dist-info`` directory.
+		Returns the parsed value of the :envvar:`SOURCE_DATE_EPOCH` environment variable, or :py:obj:`None` if unset.
 
-		:param dist_info: The absolute path to the ``*.dist-info`` directory.
+		See https://reproducible-builds.org/specs/source-date-epoch/ for the specification.
+
+		:raises ValueError: if the value is in an invalid format.
 		"""
 
-		top_level = first(posixpath.split(self.config["package"]), default=self.config["package"])
-		(dist_info / "top_level.txt").write_clean(top_level)
-		self.report_written(dist_info / "top_level.txt")
+		# If SOURCE_DATE_EPOCH is set (e.g. by Debian), it's used for timestamps inside the wheel.
+		epoch: Optional[str] = os.environ.get("SOURCE_DATE_EPOCH")
+		if epoch is None:
+			return None
+		elif epoch.isdigit():
+			return datetime.utcfromtimestamp(int(epoch))
+		else:
+			raise ValueError(f"'SOURCE_DATE_EPOCH' must be an integer with no fractional component, not {epoch!r}")
 
 	def create_wheel_archive(self) -> str:
 		"""
@@ -753,31 +764,90 @@ class WheelBuilder(AbstractBuilder):
 		wheel_filename = self.out_dir / f"{self.archive_name}-{self.tag}.whl"
 		self.out_dir.maybe_make(parents=True)
 
-		with ZipFile(wheel_filename, mode='w') as wheel_archive:
+		mtime = self.get_source_epoch()
+
+		non_record_filenames = []
+		record_filenames = []
+
+		for file in self.build_dir.rglob('*'):
+			if not file.is_file():
+				continue
+			if "RECORD" in file.name and self.dist_info.name in file.parts:
+				record_filenames.append(file)
+				continue
+
+			non_record_filenames.append(file)
+
+		record_filenames = sort_paths(*record_filenames, self.dist_info / "RECORD")
+
+		with handy_archives.ZipFile(wheel_filename, mode='w') as wheel_archive:
 			with (self.dist_info / "RECORD").open('w') as fp:
-				for file in self.build_dir.rglob('*'):
-					if "RECORD" in file.name and self.dist_info.name in file.parts:
-						continue
-					if not file.is_file():
-						continue
+				for file in sort_paths(*non_record_filenames):
 
-					fp.write(get_record_entry(file, relative_to=self.build_dir))
-					fp.write('\n')
-					wheel_archive.write(file, arcname=file.relative_to(self.build_dir))
+					fp.write(f"{get_record_entry(file, relative_to=self.build_dir)}\n")
 
-				for file in self.dist_info.rglob("RECORD*"):
-					if file.is_file():
-						fp.write(f"{file.relative_to(self.build_dir).as_posix()},,\n")
+					wheel_archive.write_file(
+							file,
+							arcname=file.relative_to(self.build_dir),
+							mtime=mtime,
+							)
 
-			for file in self.dist_info.rglob("RECORD*"):
-				if file.is_file():
-					wheel_archive.write(file, arcname=file.relative_to(self.build_dir))
-					self.report_written(file)
+				for file in record_filenames:
+					fp.write(f"{file.relative_to(self.build_dir).as_posix()},,\n")
+
+			for file in record_filenames:
+				wheel_archive.write_file(
+						file,
+						arcname=file.relative_to(self.build_dir),
+						mtime=mtime,
+						)
+				self.report_written(file)
 
 		self._echo(Fore.GREEN(f"Wheel created at {wheel_filename.resolve().as_posix()}"))
 
 		return wheel_filename.name
 
+	def create_editables_files(self) -> Iterator[ComparableRequirement]:
+		"""
+		Generate files with `editables`_ for use in a :pep:`660` wheel.
+
+		.. _editables: https://pypi.org/project/editables/
+
+		.. extras-require:: editable
+			:scope: method
+			:pyproject:
+
+		:returns: An iterator of additional runtime requirements which should be added to the wheel's ``METADATA`` file.
+		"""
+
+		# this package
+		from whey._editable import EditableProject
+
+		pkgdir = self.project_dir / self.config["source-dir"] / self.config["package"]
+
+		if not pkgdir.is_dir():
+			message = f"Package directory {self.config['package']!r} not found"
+
+			if self.config["source-dir"]:
+				raise FileNotFoundError(f"{message} in {self.config['source-dir']!r}.")
+			else:
+				raise FileNotFoundError(f"{message}.")
+
+		my_project = EditableProject(
+				self.config["name"].replace('-', '_'),
+				pkgdir.parent,
+				)
+
+		my_project.map_or_add_to_path(self.config["package"], pkgdir)
+
+		for name, content in my_project.files():
+			target = self.build_dir / name
+			target.parent.maybe_make(parents=True)
+			target.write_clean(content)
+			self.report_written(target)
+
+		yield from map(ComparableRequirement, my_project.dependencies())
+
 	def build_wheel(self) -> str:
 		"""
 		Build the binary wheel distribution.
@@ -794,11 +864,56 @@ class WheelBuilder(AbstractBuilder):
 		self.copy_additional_files()
 		self.write_license(self.dist_info)
 		self.write_entry_points()
-		self.write_metadata(self.dist_info / "METADATA")
+		self.write_metadata(self.dist_info / "METADATA", self.get_metadata_map())
 		self.write_wheel()
-		self.write_top_level(self.dist_info)
 		self.call_additional_hooks()
 
 		return self.create_wheel_archive()
 
 	build = build_wheel
+
+	def build_editable(self) -> str:
+		"""
+		Build an editable wheel.
+
+		An "editable" wheel uses the wheel format not for distribution but as ephemeral communication
+		between the build system and the front end.
+		This avoids having the build backend install anything directly.
+		This wheel must not be exposed to end users, nor cached, nor distributed.
+
+		You should use a different ``build_dir`` and ``out_dir`` to those used for standard wheel builds.
+
+		The default implementation of this method does not call
+		:meth:`~.AbstractBuilder.copy_source` or :meth:`~.AbstractBuilder.copy_additional_files`.
+
+		.. extras-require:: editable
+			:scope: method
+			:pyproject:
+
+		:return: The filename of the created archive.
+		"""
+
+		if self.build_dir.is_dir():
+			shutil.rmtree(self.build_dir)
+
+		self.build_dir.maybe_make(parents=True)
+
+		extra_deps = self.create_editables_files()
+		self.write_license(self.dist_info)
+		self.write_entry_points()
+
+		metadata = self.get_metadata_map()
+
+		for dep in extra_deps:
+			if not any(dep.name in req for req in metadata.get_all("Requires-Dist", ())):
+				# Additional runtime requirement for editable wheels.
+				metadata["Requires-Dist"] = str(dep)
+
+		# Prevents uploading to PyPI, which you shouldn't do with an editable wheel.
+		metadata["Classifier"] = "Private :: Do Not Upload"
+
+		self.write_metadata(self.dist_info / "METADATA", metadata)
+		self.write_wheel()
+		self.call_additional_hooks()
+
+		return self.create_wheel_archive()
